@@ -48,7 +48,6 @@
 #endif
 
 #include <stdio.h>
-#define NEED_EVENTS
 #include <X11/X.h>
 #include <X11/Xproto.h>
 
@@ -66,15 +65,26 @@
 #include "xf86Xinput.h"
 #include "xf86_OSproc.h"
 #include "xf86OSmouse.h"
+#include "xf86Priv.h"
 #include "compiler.h"
 
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
 #include <xserver-properties.h>
+#include "exevents.h"
+#endif
+
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 18
+#define LogMessageVerbSigSafe xf86MsgVerb
 #endif
 
 #include "xisb.h"
 #include "mipointer.h"
 
+#ifndef HAVE_XORG_SERVER_1_5_0
+#include <xf86_ansic.h>
+#include <xf86_libc.h>
+#endif
+
 /*****************************************************************************
  *	Local Headers
  ****************************************************************************/
@@ -105,31 +115,41 @@
 
 /*
  * Standard four digit version string expected by VMware Tools installer.
- * As the driver's version is only  {major, minor, patchlevel}, simply append an
- * extra zero for the fourth digit.
+ * As the driver's version is only  {major, minor, patchlevel},
+ * The fourth digit may describe the commit number relative to the
+ * last version tag as output from `git describe`
  */
 #ifdef __GNUC__
+#ifdef VMW_SUBPATCH
+const char vm_mouse_version[] __attribute__((section(".modinfo"),unused)) =
+    "version=" VMMOUSE_DRIVER_VERSION_STRING "." VMW_STRING(VMW_SUBPATCH);
+#else
 const char vm_mouse_version[] __attribute__((section(".modinfo"),unused)) =
     "version=" VMMOUSE_DRIVER_VERSION_STRING ".0";
+#endif /*VMW_SUBPATCH*/
 #endif
 
 
 /*****************************************************************************
  *	static function header
  ****************************************************************************/
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+static int VMMousePreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);
+#else
 static InputInfoPtr VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags);
+static void VMMouseCloseProc(InputInfoPtr pInfo);
+static Bool VMMouseConvertProc(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v2,
+			       int v3, int v4, int v5, int *x, int *y);
+#endif
 static void VMMouseUnInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);
 static void MouseCommonOptions(InputInfoPtr pInfo);
 static void GetVMMouseMotionEvent(InputInfoPtr pInfo);
 static void VMMousePostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy, int dz, int dw);
 static void VMMouseDoPostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy);
 static Bool VMMouseDeviceControl(DeviceIntPtr device, int mode);
-static void VMMouseCloseProc(LocalDevicePtr local);
-static int  VMMouseControlProc(LocalDevicePtr local, xDeviceCtl * control);
+static int  VMMouseControlProc(InputInfoPtr pInfo, xDeviceCtl * control);
 static void VMMouseReadInput(InputInfoPtr pInfo);
 static int  VMMouseSwitchMode(ClientPtr client, DeviceIntPtr dev, int mode);
-static Bool VMMouseConvertProc(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v2,
-			       int v3, int v4, int v5, int *x, int *y);
 static void MouseCtrl(DeviceIntPtr device, PtrCtrl *ctrl);
 
 /******************************************************************************
@@ -198,8 +218,11 @@ InputDriverRec VMMOUSE = {
    NULL,
    VMMousePreInit,
    VMMouseUnInit,
-   NULL,
+   NULL
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 11
+       ,
    0
+#endif
 };
 
 static char reverseMap[32] = { 0,  4,  2,  6,  1,  5,  3,  7,
@@ -209,6 +232,75 @@ static char reverseMap[32] = { 0,  4,  2,  6,  1,  5,  3,  7,
 
 #define reverseBits(map, b)	(((b) & ~0x0f) | map[(b) & 0x0f])
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 14
+
+static InputOption*
+input_option_new(InputOption *list, char *key, char *value)
+{
+   InputOption *new;
+
+   new = calloc(1, sizeof(InputOption));
+   new->key = key;
+   new->value = value;
+   new->next = list;
+   return new;
+}
+
+static void
+input_option_free_list(InputOption **opts)
+{
+   InputOption *tmp = *opts;
+   while(*opts)
+   {
+      tmp = (*opts)->next;
+      free((*opts)->key);
+      free((*opts)->value);
+      free((*opts));
+      *opts = tmp;
+   }
+}
+#endif
+
+static int
+VMMouseInitPassthru(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
+{
+   InputAttributes *attrs = NULL;
+   InputOption *input_options = NULL;
+   pointer options;
+   DeviceIntPtr dev;
+   int rc;
+
+   options = xf86OptionListDuplicate(pInfo->options);
+   options = xf86ReplaceStrOption(options, "Driver", "mouse");
+
+   while(options) {
+      input_options = input_option_new(input_options,
+                                       xf86OptionName(options),
+                                       xf86OptionValue(options));
+      options = xf86NextOption(options);
+   }
+
+   rc = NewInputDeviceRequest(input_options, attrs, &dev);
+
+   input_option_free_list(&input_options);
+
+   return rc;
+}
+
+#else /* if ABI_XINPUT_VERSION < 12 */
+static InputInfoPtr
+VMMouseInitPassthru(InputDriverPtr drv, IDevPtr dev, int flags)
+{
+   InputDriverRec *passthruMouse;
+   passthruMouse = (InputDriverRec *)LoaderSymbol("MOUSE");
+   if(passthruMouse != NULL) {
+      return (passthruMouse->PreInit)(drv, dev, flags);
+   } else {
+      return NULL;
+   }
+}
+#endif
 
 /*
  *----------------------------------------------------------------------
@@ -229,12 +321,14 @@ static char reverseMap[32] = { 0,  4,  2,  6,  1,  5,  3,  7,
  *----------------------------------------------------------------------
  */
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
+static int
+VMMouseNewPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);
+
 static InputInfoPtr
 VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 {
    InputInfoPtr pInfo;
-   MouseDevPtr pMse;
-   VMMousePrivPtr mPriv;
 
 #ifndef NO_MOUSE_MODULE
 {
@@ -249,15 +343,16 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 }
 #endif
 
-   mPriv = xcalloc (1, sizeof (VMMousePrivRec));
-
-
-   if (!mPriv) {
-      return NULL;
+   /*
+    * enable hardware access
+    */
+   if (!xorgHWAccess) {
+      if (xf86EnableIO())
+          xorgHWAccess = TRUE;
+      else
+          return NULL;
    }
 
-   mPriv->absoluteRequested = FALSE;
-
    /*
     * try to enable vmmouse here
     */
@@ -266,22 +361,12 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
        * vmmouse failed
        * Fall back to normal mouse module
        */
-      InputDriverRec *passthruMouse;
       xf86Msg(X_ERROR, "VMWARE(0): vmmouse enable failed\n");
-      mPriv->vmmouseAvailable = FALSE;
-      passthruMouse = (InputDriverRec *)LoaderSymbol("MOUSE");
-      xfree(mPriv);
-      if(passthruMouse != NULL){
-	 return (passthruMouse->PreInit)(drv, dev, flags);
-      } else {
-	 return NULL;
-      }
-
+      return VMMouseInitPassthru(drv, dev, flags);
    } else {
       /*
        * vmmouse is available
        */
-      mPriv->vmmouseAvailable = TRUE;
       xf86Msg(X_INFO, "VMWARE(0): vmmouse is available\n");
       /*
        * Disable the absolute pointing device for now
@@ -291,22 +376,14 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
    }
 
    if (!(pInfo = xf86AllocateInput(drv, 0))) {
-      xfree(mPriv);
       return NULL;
    }
 
-   /* Settup the pInfo */
    pInfo->name = dev->identifier;
-   pInfo->type_name = XI_MOUSE;
-   pInfo->flags = XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS;
-   pInfo->device_control = VMMouseDeviceControl;
-   pInfo->read_input = VMMouseReadInput;
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
    pInfo->motion_history_proc = xf86GetMotionEvents;
 #endif
-   pInfo->control_proc = VMMouseControlProc;
    pInfo->close_proc = VMMouseCloseProc;
-   pInfo->switch_mode = VMMouseSwitchMode;
    pInfo->conversion_proc = VMMouseConvertProc;
    pInfo->reverse_conversion_proc = NULL;
    pInfo->fd = -1;
@@ -314,11 +391,72 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
    pInfo->private_flags = 0;
    pInfo->always_core_feedback = 0;
    pInfo->conf_idev = dev;
+   pInfo->flags = XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS;
+
+   /* Collect the options, and process the common options. */
+   xf86CollectInputOptions(pInfo, NULL, NULL);
+   xf86ProcessCommonOptions(pInfo, pInfo->options);
+
+   if (VMMouseNewPreInit(drv, pInfo, flags) == Success)
+       pInfo->flags |= XI86_CONFIGURED;
+
+   return pInfo;
+}
+
+static int
+VMMouseNewPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
+#else /* if ABI_XINPUT_VERSION >= 12 */
+static int
+VMMousePreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
+#endif
+{
+   MouseDevPtr pMse = NULL;
+   VMMousePrivPtr mPriv = NULL;
+   int rc = Success;
+
+   /* Enable hardware access. */
+   if (!xorgHWAccess) {
+      if (xf86EnableIO())
+          xorgHWAccess = TRUE;
+      else {
+          rc = BadValue;
+          goto error;
+      }
+   }
+
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+   /* For ABI < 12, we need to return the wrapped driver's pInfo (see
+    * above). ABI 12, we call NIDR and are done */
+   if (!VMMouseClient_Enable()) {
+      xf86Msg(X_ERROR, "VMWARE(0): vmmouse enable failed\n");
+      return VMMouseInitPassthru(drv, pInfo, flags);
+   } else {
+      xf86Msg(X_INFO, "VMWARE(0): vmmouse is available\n");
+      VMMouseClient_Disable();
+   }
+#endif
+
+   mPriv = calloc (1, sizeof (VMMousePrivRec));
+
+   if (!mPriv) {
+      rc = BadAlloc;
+      goto error;
+   }
+
+   mPriv->absoluteRequested = FALSE;
+   mPriv->vmmouseAvailable = TRUE;
+
+   /* Settup the pInfo */
+   pInfo->type_name = XI_MOUSE;
+   pInfo->device_control = VMMouseDeviceControl;
+   pInfo->read_input = VMMouseReadInput;
+   pInfo->control_proc = VMMouseControlProc;
+   pInfo->switch_mode = VMMouseSwitchMode;
 
    /* Allocate the MouseDevRec and initialise it. */
-   if (!(pMse = xcalloc(sizeof(MouseDevRec), 1))) {
-      xfree(mPriv);
-      return pInfo;
+   if (!(pMse = calloc(sizeof(MouseDevRec), 1))) {
+      rc = BadAlloc;
+      goto error;
    }
 
    pInfo->private = pMse;
@@ -328,10 +466,6 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
    pMse->mousePriv = mPriv;
 
 
-   /* Collect the options, and process the common options. */
-   xf86CollectInputOptions(pInfo, NULL, NULL);
-   xf86ProcessCommonOptions(pInfo, pInfo->options);
-
    /* Check if the device can be opened. */
    pInfo->fd = xf86OpenSerial(pInfo->options);
    if (pInfo->fd == -1) {
@@ -339,11 +473,8 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 	 xf86Msg(X_WARNING, "%s: cannot open input device\n", pInfo->name);
       else {
 	 xf86Msg(X_ERROR, "%s: cannot open input device\n", pInfo->name);
-	 if (pMse->mousePriv)
-	    xfree(pMse->mousePriv);
-	 xfree(pMse);
-	 pInfo->private = NULL;
-	 return pInfo;
+	 rc = BadValue;
+	 goto error;
       }
    }
    xf86CloseSerial(pInfo->fd);
@@ -355,8 +486,16 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
    /* set up the current screen num */
    mPriv->screenNum = xf86SetIntOption(pInfo->options, "ScreenNumber", 0);
 
-   pInfo->flags |= XI86_CONFIGURED;
-   return pInfo;
+   return Success;
+
+error:
+   pInfo->private = NULL;
+   if (mPriv)
+      free(mPriv);
+   if (pMse)
+      free(pMse);
+
+   return rc;
 }
 
 
@@ -590,19 +729,14 @@ static void
 MouseCommonOptions(InputInfoPtr pInfo)
 {
    MouseDevPtr pMse;
-   MessageType from = X_DEFAULT;
    char *s;
-   int origButtons;
 
    pMse = pInfo->private;
 
    pMse->buttons = xf86SetIntOption(pInfo->options, "Buttons", 0);
-   from = X_CONFIG;
    if (!pMse->buttons) {
       pMse->buttons = MSE_DFLTBUTTONS;
-      from = X_DEFAULT;
    }
-   origButtons = pMse->buttons;
 
    /*
     * "emulate3Buttons" and "Drag Lock" is not supported
@@ -645,19 +779,12 @@ MouseCommonOptions(InputInfoPtr pInfo)
       }
       if (msg) {
 	 xf86Msg(X_CONFIG, "%s: ZAxisMapping: %s\n", pInfo->name, msg);
-	 xfree(msg);
+	 free(msg);
       } else {
 	 xf86Msg(X_WARNING, "%s: Invalid ZAxisMapping value: \"%s\"\n",
 		 pInfo->name, s);
       }
    }
-
-   /*
-    * Emulatewheel is not supported
-    */
-   if (origButtons != pMse->buttons)
-      from = X_CONFIG;
-
 }
 
 
@@ -678,9 +805,18 @@ MouseCommonOptions(InputInfoPtr pInfo)
  */
 
 static void
-VMMouseUnInit(InputDriverPtr drv, LocalDevicePtr local, int flags)
+VMMouseUnInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 {
+   MouseDevPtr pMse = pInfo->private;
+
    xf86Msg(X_INFO, "VMWARE(0): VMMouseUnInit\n");
+
+   if (pMse) {
+       VMMousePrivPtr mPriv = (VMMousePrivPtr)pMse->mousePriv;
+       free(mPriv);
+   }
+
+   xf86DeleteInput(pInfo, flags);
 }
 
 
@@ -708,7 +844,6 @@ VMMouseDeviceControl(DeviceIntPtr device, int mode)
 {
    InputInfoPtr pInfo;
    MouseDevPtr pMse;
-   VMMousePrivPtr mPriv;
    unsigned char map[MSE_MAXBUTTONS + 1];
    int i;
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
@@ -719,7 +854,6 @@ VMMouseDeviceControl(DeviceIntPtr device, int mode)
    pInfo = device->public.devicePrivate;
    pMse = pInfo->private;
    pMse->device = device;
-   mPriv = (VMMousePrivPtr)pMse->mousePriv;
 
    switch (mode){
    case DEVICE_INIT:
@@ -734,10 +868,10 @@ VMMouseDeviceControl(DeviceIntPtr device, int mode)
       btn_labels[0] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_LEFT);
       btn_labels[1] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_MIDDLE);
       btn_labels[2] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_RIGHT);
-      btn_labels[4] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_WHEEL_UP);
-      btn_labels[5] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_WHEEL_DOWN);
-      btn_labels[6] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_HWHEEL_LEFT);
-      btn_labels[7] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_HWHEEL_RIGHT);
+      btn_labels[3] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_WHEEL_UP);
+      btn_labels[4] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_WHEEL_DOWN);
+      btn_labels[5] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_HWHEEL_LEFT);
+      btn_labels[6] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_HWHEEL_RIGHT);
       /* other buttons are unknown */
 
 #ifdef ABS_VALUATOR_AXES
@@ -776,13 +910,21 @@ VMMouseDeviceControl(DeviceIntPtr device, int mode)
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
 				axes_labels[0],
 #endif
-				0, 65535, 10000, 0, 10000);
+				0, 65535, 10000, 0, 10000
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+                                , Absolute
+#endif
+                                );
 #else
       xf86InitValuatorAxisStruct(device, 0,
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
 				axes_labels[0],
 #endif
-				0, -1, 1, 0, 1);
+				0, -1, 1, 0, 1
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+                                , Relative
+#endif
+                                );
 #endif
       xf86InitValuatorDefaults(device, 0);
       /* Y valuator */
@@ -791,13 +933,21 @@ VMMouseDeviceControl(DeviceIntPtr device, int mode)
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
 				axes_labels[1],
 #endif
-				0, 65535, 10000, 0, 10000);
+				0, 65535, 10000, 0, 10000
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+                                , Absolute
+#endif
+                                );
 #else
       xf86InitValuatorAxisStruct(device, 1,
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
 				axes_labels[1],
 #endif
-				0, -1, 1, 0, 1);
+				0, -1, 1, 0, 1
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+                                , Relative
+#endif
+                                );
 #endif
       xf86InitValuatorDefaults(device, 1);
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
@@ -869,6 +1019,15 @@ VMMouseDeviceControl(DeviceIntPtr device, int mode)
       usleep(300000);
       break;
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) * 100 + GET_ABI_MINOR(ABI_XINPUT_VERSION) >= 1901
+   case  DEVICE_ABORT:
+      if (pInfo->fd != -1) {
+	 VMMousePrivPtr mPriv = (VMMousePrivPtr)pMse->mousePriv;
+	 if( mPriv->vmmouseAvailable )
+	    VMMouseClient_Disable();
+         break;
+      }
+#endif
    }
 
    return Success;
@@ -912,7 +1071,7 @@ VMMouseReadInput(InputInfoPtr pInfo)
        */
       VMMouseClient_RequestAbsolute();
       mPriv->absoluteRequested = TRUE;
-      xf86Msg(X_INFO, "VMWARE(0): vmmouse enable absolute mode\n");
+      LogMessageVerbSigSafe(X_INFO, -1, "VMWARE(0): vmmouse enable absolute mode\n");
    }
 
    /*
@@ -969,17 +1128,17 @@ GetVMMouseMotionEvent(InputInfoPtr pInfo){
    VMMousePrivPtr mPriv;
    int buttons, dx, dy, dz, dw;
    VMMOUSE_INPUT_DATA  vmmouseInput;
-   int ps2Buttons = 0;
    int numPackets;
 
    pMse = pInfo->private;
    mPriv = (VMMousePrivPtr)pMse->mousePriv;
    while((numPackets = VMMouseClient_GetInput(&vmmouseInput))){
+      int ps2Buttons = 0;
       if (numPackets == VMMOUSE_ERROR) {
          VMMouseClient_Disable();
          VMMouseClient_Enable();
          VMMouseClient_RequestAbsolute();
-         xf86Msg(X_INFO, "VMWARE(0): re-requesting absolute mode after reset\n");
+         LogMessageVerbSigSafe(X_INFO, -1, "VMWARE(0): re-requesting absolute mode after reset\n");
          break;
       }
 
@@ -1025,7 +1184,7 @@ GetVMMouseMotionEvent(InputInfoPtr pInfo){
  */
 
 static int
-VMMouseControlProc(LocalDevicePtr local, xDeviceCtl * control)
+VMMouseControlProc(InputInfoPtr pInfo, xDeviceCtl * control)
 {
    xf86Msg(X_INFO, "VMWARE(0): VMMouseControlProc\n");
    return (Success);
@@ -1047,11 +1206,13 @@ VMMouseControlProc(LocalDevicePtr local, xDeviceCtl * control)
  *----------------------------------------------------------------------
  */
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
 static void
-VMMouseCloseProc(LocalDevicePtr local)
+VMMouseCloseProc(InputInfoPtr pInfo)
 {
    xf86Msg(X_INFO, "VMWARE(0): VMMouseCloseProc\n");
 }
+#endif
 
 
 /*
@@ -1092,6 +1253,7 @@ VMMouseSwitchMode(ClientPtr client, DeviceIntPtr dev, int mode)
  *----------------------------------------------------------------------
  */
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
 static Bool
 VMMouseConvertProc(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v2,
 	     int v3, int v4, int v5, int *x, int *y)
@@ -1122,6 +1284,7 @@ VMMouseConvertProc(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v
    }
    return TRUE;
 }
+#endif
 
 
 #ifdef XFree86LOADER
@@ -1201,7 +1364,7 @@ VMMousePlug(pointer	module,
    if (!LoadSubModule(module, name, NULL, NULL, NULL, NULL, errmaj, errmin)) {
       LoaderErrorMsg(NULL, name, *errmaj, *errmin);
    }
-   xfree(name);
+   free(name);
 }
 #endif
 
